﻿#
# FFT options
# -----------
if(${CMAKE_PROJECT_NAME}_USE_MKL)
    list(FIND FUSE_LIST CBLAS MKL_ALREADY_FOUND)
    if(${MKL_ALREADY_FOUND} EQUAL -1)
        message(FATAL_ERROR "MKL not found for FFTW")
    else()
        set(FFTW_FOUND TRUE)
        if(TARGET MKL::MKL)
            list(APPEND FFTW_TARGET MKL::MKL)
        else()
            list(APPEND FFTW_TARGET BLAS::BLAS)
        endif()
    endif()
else()
    find_package(PkgConfig REQUIRED)
    pkg_search_module(FFTW REQUIRED fftw3 IMPORTED_TARGET)
    pkg_search_module(FFTWF REQUIRED fftw3f IMPORTED_TARGET)
    list(APPEND FFTW_TARGET PkgConfig::FFTW)
    list(APPEND FFTW_TARGET PkgConfig::FFTWF)
    set(FFTW_FOUND TRUE)
endif()

if(DEFINED FFTW_FOUND)
    list(APPEND FUSE_LIST FFTW)
else()
    message(FATAL_ERROR "FFTW has not been found, ${CMAKE_PROJECT_NAME} will continue to compile but some applications will be disabled.")
    message(FATAL_ERROR "If you have FFTW set FFTW_LIBDIR, FFTW_INCDIR or FFTW_DIR (CMake variables using -D or environment variables).")
    set(${CMAKE_PROJECT_NAME}_USE_FFT OFF)
endif()
